#include "cam_preprocess.h"

CamPreprocess::CamPreprocess(uint8_t id)
{
    this->id = id;
}

CamPreprocess::~CamPreprocess()
{
    int ret;
    ret = hi_mpi_dvpp_free(this->inputPicInfo.picture_address);
    if (!ret)
    {
        this->logger.LogInfo() << "release memory of input image";
    }
    else
    {
        this->logger.LogError() << "failed to release memory of input image with value: " << ret;
    }

    ret = hi_mpi_dvpp_free(this->outputPicInfo.picture_address);
    if (!ret)
    {
        this->logger.LogInfo() << "release memory of output image";
    }
    else
    {
        this->logger.LogError() << "failed to release memory of output image with value: " << ret;
    }

    ret = hi_mpi_dvpp_free(this->cropResizeRegion.dest_pic_info.picture_address);
    if (!ret)
    {
        this->logger.LogInfo() << "release memory of cropResize image";
    }
    else
    {
        this->logger.LogError() << "failed to release memory of cropResize image with value: " << ret;
    }

    ret = hi_mpi_vpc_destroy_chn(this->channelIDs);
    if (!ret)
    {
        this->logger.LogInfo() << "destroy channel " << this->channelIDs;
    }
    else
    {
        this->logger.LogError() << "failed to destroy channel " << this->channelIDs << " with value: " << ret;
    }

    ret = hi_mpi_sys_exit();
    if (!ret)
    {
        this->logger.LogInfo() << "Media data processing system de-initialization";
    }
    else
    {
        this->logger.LogError() << "Media data processing system de-initialization failed";
    }

    ret = aclrtDestroyContext(this->g_context);
    if (!ret)
    {
        this->logger.LogInfo() << "release Context";
    }
    else
    {
        this->logger.LogError() << "release Context failed";
    }

    ret = aclrtResetDevice(0);
    if (!ret)
    {
        this->logger.LogInfo() << "release Device";
    }
    else
    {
        this->logger.LogError() << "release Device failed";
    }

    ret = aclFinalize();
    if (!ret)
    {
        this->logger.LogInfo() << "AscendCL de-initializes";
    }
    else
    {
        this->logger.LogError() << "AscendCL de-initializes failed";
    }
}

void CamPreprocess::setDefaultRemapParam()
{
    hi_u32 size;
    std::ifstream xFile(this->remapFilePathX, std::ios::binary);
    if (!xFile.is_open())
    {
        this->logger.LogError() << "Error: Failed to open file for reading.";
        return;
    }
    xFile.seekg(0, std::ios::end);
    std::streampos fileSize = xFile.tellg();
    xFile.seekg(0, std::ios::beg);
    size_t numFloats = fileSize / sizeof(float);
    std::vector<float> xArray(numFloats);
    xFile.read(reinterpret_cast<char *>(xArray.data()), xArray.size() * sizeof(float));
    xFile.close();

    std::ifstream yFile(this->remapFilePathY, std::ios::binary);
    if (!yFile.is_open())
    {
        this->logger.LogError() << "Error: Failed to open file for reading.";
        return;
    }
    yFile.seekg(0, std::ios::end);
    std::streampos fileSizeY = yFile.tellg();
    yFile.seekg(0, std::ios::beg);
    size_t numFloatsY = fileSizeY / sizeof(float);
    std::vector<float> yArray(numFloatsY);
    yFile.read(reinterpret_cast<char *>(yArray.data()), yArray.size() * sizeof(float));
    yFile.close();
    float *firstRowArray = new float[xArray.size()];
    std::copy(xArray.begin(), xArray.end(), firstRowArray);
    float *secondRowArray = new float[yArray.size()];
    std::copy(yArray.begin(), yArray.end(), secondRowArray);
    this->mappingParameters.map1 = static_cast<void *>(firstRowArray);
    this->mappingParameters.map2 = static_cast<void *>(secondRowArray);

    this->mappingParameters.src_pic_width = this->inputImageWidth;
    this->mappingParameters.src_pic_height = this->inputImageHeight;
    this->mappingParameters.dest_pic_width = this->inputImageWidth;
    this->mappingParameters.dest_pic_height = this->inputImageHeight;
    this->mappingParameters.map_size = this->mappingParameters.dest_pic_width * this->mappingParameters.dest_pic_height * sizeof(hi_float);

    this->remappingTableInformation.src_pic_width = this->inputImageWidth;
    this->remappingTableInformation.src_pic_height = this->inputImageHeight;
    this->remappingTableInformation.dest_pic_width = this->inputImageWidth;
    this->remappingTableInformation.dest_pic_height = this->inputImageHeight;
    uint8_t failed = hi_mpi_vpc_get_lut_mem_size(this->remappingTableInformation.dest_pic_width, this->remappingTableInformation.dest_pic_height, &size);
    if (failed)
    {
        this->logger.LogError() << "hi_mpi_vpc_get_lut_mem_size failed with value: " << failed;
    }
    this->remappingTableInformation.lut_size = size;

    this->warpTransformParam.src.picture_width = this->inputImageWidth;
    this->warpTransformParam.src.picture_height = this->inputImageHeight;
    this->warpTransformParam.src.picture_format = HI_PIXEL_FORMAT_YUV_SEMIPLANAR_420;
    this->warpTransformParam.src.picture_width_stride = this->inputImageWidth;
    this->warpTransformParam.src.picture_height_stride = this->inputImageHeight;
    this->warpTransformParam.src.picture_buffer_size = this->warpTransformParam.src.picture_width_stride * this->warpTransformParam.src.picture_height_stride * 3 / 2;
    this->warpTransformParam.dst.picture_width = this->inputImageWidth;
    this->warpTransformParam.dst.picture_height = this->inputImageHeight;
    this->warpTransformParam.dst.picture_format = HI_PIXEL_FORMAT_YUV_SEMIPLANAR_420;
    this->warpTransformParam.dst.picture_width_stride = this->inputImageWidth;
    this->warpTransformParam.dst.picture_height_stride = this->inputImageHeight;
    this->warpTransformParam.dst.picture_buffer_size = this->warpTransformParam.dst.picture_width_stride * this->warpTransformParam.dst.picture_height_stride * 3 / 2;
    this->warpTransformParam.transform_config.interpolation = 0;
    this->warpTransformParam.transform_config.border_type = HI_BORDER_REPLICATE;

    this->workSpaceParam[0].func = REMAP;
    this->workSpaceParam[0].max_source_pic_width = this->inputImageWidth;
    this->workSpaceParam[0].max_source_pic_height = this->inputImageHeight;
    this->workSpaceParam[0].max_dest_pic_width = this->inputImageWidth;
    this->workSpaceParam[0].max_dest_pic_height = this->inputImageHeight;
}

void CamPreprocess::setDefaultInputPicInfo()
{
    this->inputPicInfo.picture_width = this->inputImageWidth;
    this->inputPicInfo.picture_height = this->inputImageHeight;
    this->inputPicInfo.picture_format = HI_PIXEL_FORMAT_YUV_SEMIPLANAR_420;
    this->inputPicInfo.picture_width_stride = this->inputImageWidth;
    this->inputPicInfo.picture_height_stride = this->inputImageHeight;
    this->inputPicInfo.picture_buffer_size = this->inputPicInfo.picture_width_stride * this->inputPicInfo.picture_height_stride * 3 / 2;
}

void CamPreprocess::setDefaultCropReizeRegion()
{
    this->cropResizeRegion.dest_pic_info.picture_width = this->cropWidth;
    this->cropResizeRegion.dest_pic_info.picture_height = this->cropHeight;
    this->cropResizeRegion.dest_pic_info.picture_format = HI_PIXEL_FORMAT_YUV_SEMIPLANAR_420;
    this->cropResizeRegion.dest_pic_info.picture_width_stride = this->cropWidth;
    this->cropResizeRegion.dest_pic_info.picture_height_stride = this->cropHeight;
    this->cropResizeRegion.dest_pic_info.picture_buffer_size = this->cropResizeRegion.dest_pic_info.picture_width_stride * this->cropResizeRegion.dest_pic_info.picture_height_stride * 3 / 2;
    this->cropResizeRegion.crop_region.top_offset = this->cropTopOffset;
    this->cropResizeRegion.crop_region.left_offset = this->cropLeftOffset;
    this->cropResizeRegion.crop_region.crop_width = this->inputImageWidth - this->cropResizeRegion.crop_region.left_offset * 2;
    this->cropResizeRegion.crop_region.crop_height = this->inputImageHeight - this->cropResizeRegion.crop_region.top_offset * 2;
    this->cropResizeRegion.resize_info.resize_width = this->cropWidth;
    this->cropResizeRegion.resize_info.resize_height = this->cropHeight;
    this->cropResizeRegion.resize_info.interpolation = 1;
}

void CamPreprocess::setDefaultOutputPicInfo()
{
    this->outputPicInfo.picture_width = this->cropResizeRegion.dest_pic_info.picture_width;
    this->outputPicInfo.picture_height = this->cropResizeRegion.dest_pic_info.picture_height;
    this->outputPicInfo.picture_format = HI_PIXEL_FORMAT_BGR_888;
    this->outputPicInfo.picture_width_stride = this->cropResizeRegion.dest_pic_info.picture_width * 3;
    this->outputPicInfo.picture_height_stride = this->cropResizeRegion.dest_pic_info.picture_height;
    this->outputPicInfo.picture_buffer_size = this->outputPicInfo.picture_width_stride * this->outputPicInfo.picture_height_stride;
}

void CamPreprocess::setupPipelines(boost::property_tree::ptree &config, uint8_t startCameraId)
{
    const std::string key = "cam_" + std::to_string(this->id + startCameraId);

    this->inputImageWidth = config.get<unsigned int>(key + ".inputImageWidth");
    this->inputImageHeight = config.get<unsigned int>(key + ".inputImageHeight");
    this->cropWidth = config.get<unsigned int>(key + ".outputImageWidth");
    this->cropHeight = config.get<unsigned int>(key + ".outputImageHeight");
    this->cropTopOffset = config.get<unsigned int>(key + ".cropTopOffset");
    this->cropLeftOffset = config.get<unsigned int>(key + ".cropLeftOffset");
    this->encoderGopNumber = config.get<unsigned int>(key + ".encoderGopNumber");
    if (config.get_child_optional(key + ".remapFilePathX"))
    {
        this->remapFilePathX = config.get<std::string>(key + ".remapFilePathX");
        if (config.get_child_optional(key + ".remapFilePathY"))
        {
            this->remapFilePathY = config.get<std::string>(key + ".remapFilePathY");
            this->undistortion = true;
        }
        else
        {
            this->logger.LogWarn() << "The INI file does not contain the key 'remapFilePathY', so undistortion part is disabled.";
            this->undistortion = false;
        }
    }
    else
    {
        this->logger.LogWarn() << "The INI file does not contain the key 'remapFilePathX', so undistortion part is disabled.";
        this->undistortion = false;
    }
}

void CamPreprocess::init(boost::property_tree::ptree &config, uint8_t startCameraId)
{
    uint8_t failed;
    this->setupPipelines(config, startCameraId);
    failed = this->initACL();
    this->initMPI();
    if (!failed)
    {
        this->logger.LogInfo() << "initialize pipeline done";
    }
    else
    {
        this->logger.LogError() << "initialize pipeline failed with result = 0x" << failed;
    }
    this->createEncoderChannel(this->inputImageWidth, this->inputImageHeight, 5);
}

uint8_t CamPreprocess::initACL()
{
    aclError aclRet;
    aclRet = aclrtGetRunMode(&this->runMode);
    if (!aclRet)
    {
        this->logger.LogInfo() << "ACL get run mode successfully";
    }
    else
    {
        this->logger.LogError() << "ACL didn't get run mode successfully, ret = 0x" << aclRet;
    }
    aclRet = aclInit(nullptr);
    if (!aclRet)
    {
        this->logger.LogInfo() << "ACL Initialized successfully";
    }
    else
    {
        this->logger.LogError() << "ACL didn't initialized successfully, ret = 0x" << aclRet;
    }
    aclRet = aclrtSetDevice(0);
    if (!aclRet)
    {
        this->logger.LogInfo() << "ACL set device successfully";
    }
    else
    {
        this->logger.LogError() << "ACL didn't set device successfully, ret = 0x" << aclRet;
    }
    aclRet = aclrtCreateContext(&this->g_context, 0);
    if (!aclRet)
    {
        this->logger.LogInfo() << "ACL context successfully";
    }
    else
    {
        this->logger.LogError() << "ACL didn't context successfully, ret = 0x" << aclRet;
    }
    return 0;
}

int32_t CamPreprocess::getResizedCroppedImage()
{
    int32_t failed = 0;
    uint32_t taskID = 0;
    if (this->undistortion)
    {
        failed = hi_mpi_vpc_crop_resize(this->channelIDs, &this->warpTransformParam.dst, &this->cropResizeRegion, 1, &taskID, -1);
    }
    else
    {
        failed = hi_mpi_vpc_crop_resize(this->channelIDs, &this->inputPicInfo, &this->cropResizeRegion, 1, &taskID, -1);
    }
    if (!failed)
    {
        failed = hi_mpi_vpc_get_process_result(this->channelIDs, taskID, -1);
        if (failed)
        {
            this->logger.LogError() << "failed to get cropped and resized image with result 0x" << failed;
        }
    }
    else
    {
        this->logger.LogError() << "failed to crop and resize with result 0x" << failed;
    }
    return failed;
}

int32_t CamPreprocess::getConvertedImage()
{
    int32_t failed = 0;
    uint32_t taskID = 0;
    failed = hi_mpi_vpc_convert_color(this->channelIDs, &this->cropResizeRegion.dest_pic_info, &this->outputPicInfo, &taskID, -1);
    if (!failed)
    {
        failed = hi_mpi_vpc_get_process_result(this->channelIDs, taskID, -1);
        if (failed)
        {
            this->logger.LogError() << "failed to get converted image with result 0x" << failed;
        }
    }
    else
    {
        this->logger.LogError() << "failed to convert with result 0x" << failed;
    }
    return failed;
}

int32_t CamPreprocess::getUndistortedImage()
{
    int32_t failed = 0;
    uint32_t taskID = 0;
    failed = hi_mpi_vpc_lut_remap(this->channelIDs, &this->warpTransformParam, &this->remappingTableInformation, &taskID, -1);
    if (!failed)
    {
        failed = hi_mpi_vpc_get_process_result(this->channelIDs, taskID, -1);
        if (failed)
        {
            this->logger.LogError() << "failed to get undistorted image with result 0x" << failed;
        }
    }
    else
    {
        this->logger.LogError() << "failed to undistort with result 0x" << failed;
    }
    return failed;
}

int32_t CamPreprocess::createVPCChannel()
{
    int32_t failed = 0;
    failed = hi_mpi_vpc_sys_create_chn(&this->channelIDs, &this->channelAttr);
    if (failed)
    {
        return failed;
    }
    else
    {
        this->logger.LogInfo() << "channel " << this->channelIDs << " created successfully";
    }
    return failed;
}

uint8_t CamPreprocess::allocateByPicInfo(hi_vpc_pic_info &picInfo)
{
    int32_t failed = 0;
    failed = hi_mpi_dvpp_malloc(0, &picInfo.picture_address, picInfo.picture_buffer_size);
    if (failed)
    {
        this->logger.LogError() << "failed to allocate dvpp memory for channel with result 0x" << failed;
    }
    return failed;
}

uint8_t CamPreprocess::allocateByCropResizeRegion(hi_vpc_crop_resize_region &cropResizeRegion)
{
    int32_t failed = 0;
    failed = hi_mpi_dvpp_malloc(0, &cropResizeRegion.dest_pic_info.picture_address, cropResizeRegion.dest_pic_info.picture_buffer_size);
    if (failed)
    {
        this->logger.LogError() << "failed to allocate dvpp memory for crop resize region in channel with result 0x" << failed;
    }
    return failed;
}

uint8_t CamPreprocess::allocateByUndistortion(hi_warp_transform_param &warpTransformParam, hi_remap_lut &remappingTableInformation)
{
    int32_t failed = 0;
    failed = hi_mpi_dvpp_malloc(0, &warpTransformParam.src.picture_address, warpTransformParam.src.picture_buffer_size);
    if (failed)
    {
        this->logger.LogError() << "failed to allocate dvpp memory for undistortion input in channel with result 0x" << failed;
    }
    else
    {
        failed = hi_mpi_dvpp_malloc(0, &warpTransformParam.dst.picture_address, warpTransformParam.dst.picture_buffer_size);
        if (failed)
        {
            this->logger.LogError() << "failed to allocate dvpp memory for undistortion output in channel with result 0x" << failed;
        }
        else
        {
            failed = hi_mpi_dvpp_malloc(0, &remappingTableInformation.lut, remappingTableInformation.lut_size);
        }
    }

    // move
    failed = hi_mpi_vpc_set_chn_workspace(this->channelIDs, this->workSpaceParam, 1);
    if (failed)
    {
        this->logger.LogError() << "applied for temporary cache failed with result = 0x" << failed;
    }
    else
    {
        this->logger.LogInfo() << "channel " << this->channelIDs << " applied for temporary cache successfully";
    }
    failed = hi_mpi_vpc_get_remap_lut(&this->mappingParameters, &this->remappingTableInformation);
    if (failed)
    {
        this->logger.LogError() << "failed to get remap lut with result 0x" << failed;
    }
    return failed;
}

uint32_t CamPreprocess::allocateMemory()
{
    uint8_t failed;
    failed = this->allocateByPicInfo(this->inputPicInfo);
    if (!failed)
    {
        this->logger.LogInfo() << "input memory allocated successfully";
        failed = this->allocateByPicInfo(this->outputPicInfo);
        if (!failed)
        {
            this->logger.LogInfo() << "output memory allocated successfully";
            failed = this->allocateByCropResizeRegion(this->cropResizeRegion);
            if (!failed)
            {
                this->logger.LogInfo() << "crop and resize memory allocated successfully";
                if (this->undistortion)
                {
                    failed = this->allocateByUndistortion(this->warpTransformParam, this->remappingTableInformation);
                    if (!failed)
                    {
                        this->logger.LogInfo() << "undistortion memory allocated successfully for channel";
                    }
                    else
                    {
                        this->logger.LogError() << "undistortion memory allocated failed for channel with result: " << failed;
                    }
                }
            }
            else
            {
                this->logger.LogError() << "crop and resize memory allocated failed for channel with result: " << failed;
            }
        }
        else
        {
            this->logger.LogError() << "output memory allocated failed for channel with result: " << failed;
        }
    }
    else
    {
        this->logger.LogError() << "input memory allocated failed for channel with result: " << failed;
    }
    return failed;
}

uint8_t CamPreprocess::initMPI()
{
    int32_t failed = hi_mpi_sys_init();

    if (!failed)
    {
        this->channelAttr.pic_width = 0;
        this->channelAttr.pic_height = 0;
        this->setDefaultInputPicInfo();
        this->setDefaultCropReizeRegion();
        this->setDefaultOutputPicInfo();
        if (this->undistortion)
        {
            this->setDefaultRemapParam();
        }
        failed = this->createVPCChannel();
        if (!failed)
        {
            this->logger.LogInfo() << "VPC channels created successfully";
        }
        else
        {
            this->logger.LogFatal() << "VPC channels created failed with result: " << failed;
        }
    }
    else
    {
        this->logger.LogFatal() << "mpi sys init failed with result: " << failed;
    }
    return failed;
}

int32_t CamPreprocess::createEncoderChannel(int width, int height, int bufferSize)
{
    int32_t ret;

    // set venc parameter
    hi_venc_mod_param mod_param{};
    mod_param.mod_type = HI_VENC_MOD_H264;
    hi_mpi_venc_get_mod_param(&mod_param);
    mod_param.h264_mod_param.one_stream_buf = 1;
    hi_mpi_venc_set_mod_param(&mod_param);

    // Create channel
    hi_venc_chn chn = this->id;
    hi_venc_chn_attr attr{};
    attr.venc_attr.type = HI_PT_H264;
    attr.venc_attr.profile = 0;
    attr.venc_attr.max_pic_width = width;
    attr.venc_attr.max_pic_height = height;
    attr.venc_attr.pic_width = width;
    attr.venc_attr.pic_height = height;
    attr.venc_attr.buf_size = bufferSize * 1024 * 1024;
    attr.venc_attr.is_by_frame = HI_TRUE;
    attr.rc_attr.rc_mode = HI_VENC_RC_MODE_H264_VBR;
    attr.rc_attr.h264_vbr.gop = this->encoderGopNumber;
    attr.rc_attr.h264_vbr.stats_time = 1;
    attr.rc_attr.h264_vbr.src_frame_rate = 30;
    attr.rc_attr.h264_vbr.dst_frame_rate = 30;
    attr.rc_attr.h264_vbr.max_bit_rate = 4000;
    attr.gop_attr.gop_mode = HI_VENC_GOP_MODE_NORMAL_P;
    attr.gop_attr.normal_p.ip_qp_delta = 3;
    ret = hi_mpi_venc_create_chn(chn, &attr);
    this->logger.LogInfo() << "encoder create channel returned 0x" << ret;

    hi_venc_start_param recv_param{};
    recv_param.recv_pic_num = -1;
    hi_mpi_venc_start_chn(chn, &recv_param);

    // 1.Create EPOLL instance
    this->epollFd = 0;
    this->fd = hi_mpi_venc_get_fd(chn);
    ret = hi_mpi_sys_create_epoll(10, &this->epollFd);
    this->event.events = HI_DVPP_EPOLL_IN;
    this->event.data = (void *)(unsigned long)(this->fd);
    ret = hi_mpi_sys_ctl_epoll(this->epollFd, HI_DVPP_EPOLL_CTL_ADD, this->fd, &this->event);

    return ret;
}

uint8_t CamPreprocess::h265Encoder(std::shared_ptr<hi_venc_stream> *out_enc)
{
    int32_t ret = 0;

    // If the operating mode is ACL_HOST, you need to apply for Host memory, read the input data into the Host memory, and then transfer the Host data to the Device through the aclrtMemcpy interface. After the data transmission is completed, the Host memory needs to be released in time; otherwise, the input data will be directly Read into Device memory
    if (this->runMode == ACL_HOST)
    {
        // not supported in sdk 007
        this->logger.LogError() << "ACL_HOST mode is not supported";
    }
    else
    {

        // Send input data and start encoding
        hi_video_frame_info frame{};
        frame.mod_id = HI_ID_VENC;
        frame.v_frame.width = this->inputImageWidth;
        frame.v_frame.height = this->inputImageHeight;
        frame.v_frame.field = HI_VIDEO_FIELD_FRAME;
        frame.v_frame.pixel_format = HI_PIXEL_FORMAT_YUV_SEMIPLANAR_420;
        frame.v_frame.video_format = HI_VIDEO_FORMAT_LINEAR;
        frame.v_frame.compress_mode = HI_COMPRESS_MODE_NONE;
        frame.v_frame.dynamic_range = HI_DYNAMIC_RANGE_SDR8;
        frame.v_frame.color_gamut = HI_COLOR_GAMUT_BT709;
        frame.v_frame.width_stride[0] = this->inputImageWidth;
        frame.v_frame.width_stride[1] = this->inputImageWidth;
        frame.v_frame.width_stride[2] = this->inputImageWidth;
        if (this->undistortion)
        {
            frame.v_frame.virt_addr[0] = this->warpTransformParam.src.picture_address;
        }
        else
        {
            frame.v_frame.virt_addr[0] = this->inputPicInfo.picture_address;
        }
        frame.v_frame.virt_addr[1] = (hi_void *)((uintptr_t)frame.v_frame.virt_addr[0] + (this->inputImageWidth * this->inputImageHeight));
        frame.v_frame.virt_addr[2] = (hi_void *)((uintptr_t)frame.v_frame.virt_addr[2] + 1);
        this->frame_ref += 2;
        frame.v_frame.frame_flag = 0;
        frame.v_frame.time_ref = this->frame_ref;
        frame.v_frame.pts = this->cam_data.auxInfo.ispTimeStamp.timesec * 1000000000L + this->cam_data.auxInfo.ispTimeStamp.timensec;

        ret = hi_mpi_venc_send_frame(this->id, &frame, -1);
        if (ret == 0)
        {
            // Get the encoding result
            int32_t eventCount = 0;
            // Before the encoding is completed, it will time out and block here. Once it is completed, it will continue to execute.
            ret = hi_mpi_sys_wait_epoll(this->epollFd, &this->event, 3, 1000, &eventCount);
            // 2 Get encoding results
            hi_venc_chn_status stat;
            ret = hi_mpi_venc_query_status(this->id, &stat);
            auto temp = this->id;
            if (ret == 0)
            {
                hi_venc_pack *pack_ptr = new hi_venc_pack[stat.cur_packs];
                auto enc_stream_ptr = std::shared_ptr<hi_venc_stream>(
                    new hi_venc_stream, [temp, pack = pack_ptr](auto *ptr)
                    {
                            if (0 != hi_mpi_venc_release_stream(temp, ptr)) {
                                std::cout << "hi_mpi_venc_release_stream failed" << std::endl;
                            }
                            delete[] pack;
                            delete ptr; });

                enc_stream_ptr->pack_cnt = stat.cur_packs;
                enc_stream_ptr->pack = pack_ptr;
                hi_mpi_venc_get_stream(temp, enc_stream_ptr.get(), -1);
                *out_enc = enc_stream_ptr;
            }
        }
    }
    return (uint8_t)ret;
}

int32_t CamPreprocess::runEncoderPipeline(std::shared_ptr<hi_venc_stream> *out_enc)
{
    int32_t failed = this->h265Encoder(out_enc);
    if (!failed)
    {
        this->logger.LogDebug() << "image encoded to H.265 successfully";
    }
    else
    {
        this->logger.LogError() << "image H.265 encoding error with result: 0x" << failed;
    }
    return failed;
}

int32_t CamPreprocess::runPerceptionPipeline()
{
    int32_t failed;
    if (this->undistortion)
    {
        failed = this->getUndistortedImage();
        if (!failed)
        {
            this->logger.LogDebug() << "image undistorted successfully";
        }
        else
        {
            this->logger.LogError() << "image undistortion error with result: 0x" << failed;
            return failed;
        }
    }
    failed = this->getResizedCroppedImage();
    if (!failed)
    {
        this->logger.LogDebug() << "image cropped and resized successfully";
        failed = this->getConvertedImage();
        if (!failed)
        {
            this->logger.LogDebug() << "image converted to BGR successfully";
        }
        else
        {
            this->logger.LogError() << "image converting to BGR error with result: 0x" << failed;
            return failed;
        }
    }
    else
    {
        this->logger.LogError() << "image cropping and resizing error with result: 0x" << failed;
        return failed;
    }
    return failed;
}

void **CamPreprocess::getInputAddress()
{
    if (this->undistortion)
    {
        return &this->warpTransformParam.src.picture_address;
    }
    else
    {
        return &this->inputPicInfo.picture_address;
    }
}